from .meter import Meter


class Monitor(Meter):

    def __init__(self, *args, **kwargs):
        super(Monitor, self).__init__(*args, **kwargs)

    def monitor(self, rms):
        """Extra monitor actions with RMS values

        Custom trigger conditions and actions can be written here as long as
        `action=None` is passed to the Monitor(). The Monitor instance will
        only execute monitor() as action is disabled.
        """
        pass

    # The following hooks can be used with as long as
    # `action=None` is passed to the Monitor()
    def prepopen(self):
        """Extra code before executing the script"""
        pass

    def postpopen(self):
        """Extra code after executing the script"""
        pass

    # The following hooks can be used in all cases
    def prestop(self):
        """Extra code before stop"""
        pass

    def poststop(self):
        """Extra code after stop"""
        pass



# Default settings
import os
import pyaudio


PROG = 'soundmeter'
FRAMES_PER_BUFFER = 2048
FORMAT = pyaudio.paInt16
CHANNELS = 2
RATE = 44100
AUDIO_SEGMENT_LENGTH = 0.5
USER_DIR = os.path.join(os.path.expanduser('~'), '.' + PROG)
USER_LOGFILE = os.path.join(USER_DIR, 'log')
USER_CONFIG = os.path.join(USER_DIR, 'config')
USER_SCRIPT = os.path.join(USER_DIR, 'trigger.sh')



from ctypes import * # NOQA
from contextlib import contextmanager
import os
import stat


def get_file_path(f):
    if f:
        name = getattr(f, 'name')
        if name:
            path = os.path.abspath(name)
            return path


def create_executable(path, content):
    with open(path, 'w') as f:
        f.write(content)
    s = os.stat(path)
    os.chmod(path, s.st_mode | stat.S_IEXEC)


# Work-around on error messages by alsa-lib
# http://stackoverflow.com/questions/7088672/
ERROR_HANDLER_FUNC = CFUNCTYPE(None, c_char_p, c_int,
                               c_char_p, c_int, c_char_p)


def py_error_handler(filename, line, function, err, fmt):
    pass


c_error_handler = ERROR_HANDLER_FUNC(py_error_handler)


@contextmanager
def noalsaerr():
    asound = cdll.LoadLibrary('libasound.so')
    asound.snd_lib_error_set_handler(c_error_handler)
    yield
    asound.snd_lib_error_set_handler(None)



import daemon
import logging
import pyaudio
import pydub
import wave
import signal
import subprocess
import sys
import time
import warnings
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from .settings import (FRAMES_PER_BUFFER, FORMAT, CHANNELS, RATE,
                       AUDIO_SEGMENT_LENGTH)
from .command import get_meter_kwargs, setup_user_dir
from .utils import noalsaerr


warnings.filterwarnings("ignore", category=DeprecationWarning)
_soundmeter = None


class Meter(object):

    class StopException(Exception):
        pass

    def __new__(cls, *args, **kwargs):
        if kwargs.get('daemonize'):
            with daemon.DaemonContext():
                instance = object.__new__(cls, *args, **kwargs)
                return instance
        return object.__new__(cls, *args, **kwargs)

    def __init__(self, collect=False, seconds=None, action=None,
                 threshold=None, num=None, script=None, log=None,
                 daemonize=False, verbose=False, segment=None):
        """
        :param collect: A boolean indicating whether collecting RMS values
        :param seconds: A float representing number of seconds to run the
            meter (None for forever)
        :param action: The action type ('stop', 'exec-stop' or 'exec')
        :param threshold: A string representing threshold and bound type (e.g.
            '+252', '-144')
        :param num: An integer indicating how many consecutive times the
            threshold is reached before triggering the action
        :param script: File object representing the script to be executed
        :param log: File object representing the log file
        :param daemonize: A boolean indicating whether meter is run as daemon
        :param verbose: A boolean for verbose mode
        :param segment: A float representing `AUDIO_SEGMENT_LENGTH`
        """

        global _soundmeter
        _soundmeter = self  # Register this meter globally
        self.output = StringIO()
        with noalsaerr():
            self.audio = pyaudio.PyAudio()
        self.stream = self.audio.open(format=FORMAT,
                                      channels=CHANNELS,
                                      rate=RATE,
                                      input=True,
                                      frames_per_buffer=FRAMES_PER_BUFFER)
        self.collect = collect
        self.seconds = seconds
        self.action = action
        self.threshold = threshold
        self.num = num
        self.script = script
        self.log = log
        self.daemonize = daemonize
        self.verbose = verbose
        self.segment = segment
        self.is_running = False
        self._graceful = False  # Graceful stop switch
        self._timeout = False
        self._timer = None
        self._data = {}
        self._setup_logging()

    def record(self):
        """Record PyAudio stream into StringIO output"""

        frames = []
        self.stream.start_stream()
        for i in xrange(self.num_frames):
            data = self.stream.read(FRAMES_PER_BUFFER)
            frames.append(data)
        self.stream.stop_stream()
        self.output.seek(0)
        w = wave.open(self.output, 'wb')
        w.setnchannels(CHANNELS)
        w.setsampwidth(self.audio.get_sample_size(FORMAT))
        w.setframerate(RATE)
        w.writeframes(b''.join(frames))

    def start(self):
        segment = self.segment or AUDIO_SEGMENT_LENGTH
        self.num_frames = int(RATE / FRAMES_PER_BUFFER * segment)
        if self.seconds:
            signal.setitimer(signal.ITIMER_REAL, self.seconds)
        if self.verbose:
            self._timer = time.time()
        if self.collect:
            print 'Collecting RMS values...'
        if self.action:
            # Interpret threshold
            self.get_threshold()

        try:
            self.is_running = True
            while not self._graceful:
                self.record()  # Record stream in `AUDIO_SEGMENT_LENGTH' long
                data = self.output.getvalue()
                segment = pydub.AudioSegment(data)
                rms = segment.rms
                if self.collect:
                    self.collect_rms(rms)
                self.meter(rms)
                if self.action:
                    if self.is_triggered(rms):
                        self.execute()
                self.monitor(rms)
            self.is_running = False
            self.stop()

        except self.__class__.StopException:
            self.is_running = False
            self.stop()

    def meter(self, rms):
        if not self._graceful:
            sys.stdout.write('\r%10d  ' % rms)
            sys.stdout.flush()
            if self.log:
                self.logging.info(rms)

    def graceful(self):
        """Graceful stop so that while loop in start() will stop after the
         current recording cycle"""
        self._graceful = True

    def timeout(self):
        msg = 'Timeout'
        print msg
        if self.log:
            self.logging.info(msg)
        self.graceful()

    def stop(self):
        """Stop the stream and terminate PyAudio"""
        self.prestop()
        if not self._graceful:
            self._graceful = True
        self.stream.stop_stream()
        self.audio.terminate()
        msg = 'Stopped'
        self.verbose_info(msg, log=False)
        # Log 'Stopped' anyway
        if self.log:
            self.logging.info(msg)
        if self.collect:
            if self._data:
                print 'Collected result:'
                print '    min: %10d' % self._data['min']
                print '    max: %10d' % self._data['max']
                print '    avg: %10d' % int(self._data['avg'])
        self.poststop()

    def get_threshold(self):
        """Get and validate raw RMS value from threshold"""

        if self.threshold.startswith('+'):
            if self.threshold[1:].isdigit():
                self._threshold = int(self.threshold[1:])
                self._upper = True
        elif self.threshold.startswith('-'):
            if self.threshold[1:].isdigit():
                self._threshold = int(self.threshold[1:])
                self._upper = False
        else:
            if self.threshold.isdigit():
                self._threshold = int(self.threshold)
                self._upper = True
        if not hasattr(self, '_threshold'):
            raise ValueError('Invalid threshold')

    def is_triggered(self, rms):
        if self._upper and rms > self._threshold \
                or not self._upper and rms < self._threshold:
            if 'triggered' in self._data:
                self._data['triggered'] += 1
            else:
                self._data['triggered'] = 1
        else:
            if 'triggered' in self._data:
                del self._data['triggered']
        if self._data.get('triggered') >= self.num:
            return True
        return False

    def execute(self):
        if self.action == 'stop':
            msg = 'Stop Action triggered'
            print msg
            if self.log:
                self.logging.info(msg)
            raise self.__class__.StopException('stop')

        elif self.action == 'exec-stop':
            msg = 'Exec-Stop Action triggered'
            print msg
            if self.log:
                self.logging.info(msg)
            v = 'Executing %s' % self.script
            self.verbose_info(v)
            self.popen()
            raise self.__class__.StopException('exec-stop')

        elif self.action == 'exec':
            msg = 'Exec Action triggered'
            print msg
            if self.log:
                self.logging.info(msg)
            v = 'Executing %s' % self.script
            self.verbose_info(v)
            self.popen()

    def popen(self):
        self.prepopen()
        if self.script:
            try:
                subprocess.Popen([self.script])
            except OSError, e:
                msg = 'Cannot execute the shell script: %s' % e
                print msg
                if self.log:
                    self.logging.info(msg)
        self.postpopen()

    def collect_rms(self, rms):
        """Collect and calculate min, max and average RMS values"""
        if self._data:
            self._data['min'] = min(rms, self._data['min'])
            self._data['max'] = max(rms, self._data['max'])
            self._data['avg'] = float(rms + self._data['avg']) / 2
        else:
            self._data['min'] = rms
            self._data['max'] = rms
            self._data['avg'] = rms

    def verbose_info(self, msg, log=True):
        if self.verbose:
            print msg
            if self.log and log:
                self.logging.info(msg)

    def _setup_logging(self):
        if self.log:
            self.logging = logging.basicConfig(
                filename=self.log, format='%(asctime)s %(message)s',
                level=logging.INFO)
            self.logging = logging.getLogger(__name__)

    def monitor(self, rms):
        """This function is to be overridden"""
        pass

    def prepopen(self):
        """Pre-popen hook"""
        pass

    def postpopen(self):
        """Post-popen hook"""
        pass

    def prestop(self):
        """Pre-stop hook"""
        pass

    def poststop(self):
        """Post-stop hook"""
        pass

    def __repr__(self):
        u = self.action if self.action else 'no-action'
        return '<%s: %s>' % (self.__class__.__name__, u)


def main():
    setup_user_dir()
    kwargs = get_meter_kwargs()
    """
    if kwargs['daemonize']:
        with daemon.DaemonContext():
            m = Meter(**kwargs)
            m.start()
    else:
        m = Meter(**kwargs)
        m.start()
    """
    m = Meter(**kwargs)
    m.start()


# Signal handlers
def sigint_handler(signum, frame):
    sys.stdout.write('\n')
    _soundmeter.graceful()


def sigalrm_handler(signum, frame):
    _soundmeter.timeout()


# Register signal handlers
signal.signal(signal.SIGINT, sigint_handler)
signal.signal(signal.SIGALRM, sigalrm_handler)



import os
import signal
import subprocess
from unittest import TestCase
from scripttest import TestFileEnvironment
from .command import setup_user_dir
from .meter import Meter
from .utils import create_executable


def create_run_script():
    d = os.path.dirname(__file__)
    project_path = os.path.abspath(os.path.join(d, os.pardir))
    run_script = os.path.join(project_path, 'run.py')
    content = '#!/usr/bin/env python\n'
    content += 'from soundmeter.meter import main\n\n\n'
    content += 'main()'
    if not os.path.exists(run_script):
        create_executable(run_script, content)


setup_user_dir()
create_run_script()


class TestMeter(TestCase):
    """Test Meter class programmatically"""
    def setUp(self):
        self.meter = Meter(seconds=2.0)

    def test_running(self):
        self.assertFalse(self.meter.is_running)
        self.meter.start()
        self.assertFalse(self.meter.is_running)


class TestBasicCommands(TestCase):
    """Test basic command-line invoke of the program"""
    def setUp(self):
        self.env = TestFileEnvironment('./test-output')

    def test_default(self):
        res = self.env.run('../run.py', '-s', '1')
        assert 'Timeout' in res.stdout
        self.assertEqual(res.returncode, 0)

    def test_collect(self):
        res = self.env.run('../run.py', '-s', '1', '-c')
        assert 'Collecting' in res.stdout
        self.assertEqual(res.returncode, 0)

    def test_log(self):
        res = self.env.run('../run.py', '-s', '1', '--log', 'log.txt')
        assert 'Timeout' in res.stdout
        assert 'log.txt' in res.files_created
        self.assertEqual(res.returncode, 0)

    def test_segment(self):
        res = self.env.run('../run.py', '-s', '1', '--segment', '0.2')
        assert 'Timeout' in res.stdout
        self.assertEqual(res.returncode, 0)

    def tearDown(self):
        pass


class TestCommands(TestCase):
    def test_sigint(self):
        popen = subprocess.Popen(['./run.py'])
        os.kill(popen.pid, signal.SIGINT)

    def test_arguments(self):
        popen = subprocess.Popen(['./run.py', '-t', '10000', '-a', 'stop'])
        os.kill(popen.pid, signal.SIGINT)

    def test_daemon(self):
        popen = subprocess.Popen(['./run.py', '-d'])
        os.kill(popen.pid, signal.SIGINT)



from setuptools import setup, find_packages
from soundmeter import __version__


setup(
    name='soundmeter',
    version=__version__,
    description="Simple real-time sound meter.",
    long_description=open('README.rst').read(),
    keywords='soundmeter',
    author='Shichao An',
    author_email='shichao.an@nyu.edu',
    url='https://github.com/shichao-an/soundmeter',
    license='BSD',
    install_requires=open('requirements.txt').read().splitlines(),
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'soundmeter = soundmeter.meter:main',
        ],
    },
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
    ],
)



import argparse
import os
import sys
from .settings import PROG, USER_LOGFILE, USER_SCRIPT, USER_DIR
from .utils import get_file_path, create_executable


def parse_args():
    parser = argparse.ArgumentParser(usage='%(prog)s [options]',
                                     prog=PROG)
    parser.add_argument('-c', '--collect', action='store_true',
                        help='collect RMS values to determine thresholds')
    seconds_help = 'time in seconds to run the meter (default forever)'
    parser.add_argument('-s', '--seconds', type=float,
                        help=seconds_help)
    parser.add_argument('-a', '--action',
                        choices=['stop', 'exec-stop', 'exec'],
                        help="triggered action")
    trigger_help = 'trigger condition (threshold RMS and number of times)'
    parser.add_argument('-t', '--trigger', nargs='+',
                        metavar=('[+|-]THRESHOLD', 'NUM'),
                        help=trigger_help)
    parser.add_argument('-e', '--exec', dest='script',
                        metavar='SCRIPT', type=argparse.FileType('r'),
                        default=USER_SCRIPT,
                        help='shell script to execute upon trigger')
    parser.add_argument('-d', '--daemonize', action='store_true',
                        help='run the meter in the background')
    parser.add_argument('--log', nargs='?', metavar='LOGFILE',
                        type=argparse.FileType('a'),
                        const=USER_LOGFILE,
                        help='log the meter (default to ~/.soundmeter/log)')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='verbose mode')
    segment_help = 'audio segment length recorded in seconds (defaults to 0.5)'
    parser.add_argument('--segment', metavar='SECONDS', help=segment_help)

    # Extra validation of arguments
    args = parser.parse_args()
    if args.collect:
        if args.action or args.trigger:
            msg = ('-c/--collect should not be used with -a/--action '
                   'or -t/--trigger')
            raise parser.error(msg)
    if args.segment:
        try:
            segment = float(args.segment)
        except ValueError:
            msg = '--segment must be an integer or float'
            raise parser.error(msg)
        if segment < 0.05:
            msg = '--segment cannot be smaller than 0.05'
            raise parser.error(msg)
    if args.action:
        if not args.trigger:
            msg = 'must specify -t/--trigger when using -a/--action'
            raise parser.error(msg)
        if args.action == 'stop':
            args.script = None
        if len(args.trigger) == 1:
            args.trigger.append('1')
        elif len(args.trigger) > 2:
            trigger_msg = '-t/--trigger accepts at most two arguments'
            raise parser.error(trigger_msg)
        else:
            trigger_msg = ('the second argument NUM to -t/--trigger must be an'
                           ' positive integer')
            if not args.trigger[1].isdigit():
                raise parser.error(trigger_msg)
            if args.trigger[1].isdigit() and int(args.trigger[1]) == 0:
                raise parser.error(trigger_msg)

    elif args.trigger:
        msg = 'must specify -a/--action when using -t/--trigger'
        raise parser.error(msg)

    elif args.script:
        if '-e' in sys.argv or '--exec' in sys.argv:
            msg = ('must specify -a/--action and -t/--trigger when using '
                   '-e/--exec')
            raise parser.error(msg)
        else:
            args.script = None
    return args


def get_meter_kwargs():
    kwargs = dict(parse_args()._get_kwargs())
    # Convert `trigger' into `threshold' and `num'
    if kwargs['trigger'] is not None:
        kwargs['threshold'] = kwargs['trigger'][0]
        kwargs['num'] = int(kwargs['trigger'][1])
    del kwargs['trigger']
    kwargs['script'] = get_file_path(kwargs['script'])
    kwargs['log'] = get_file_path(kwargs['log'])
    kwargs['segment'] = float(kwargs['segment']) \
        if kwargs['segment'] is not None else None
    return kwargs


def setup_user_dir():
    if not os.path.exists(USER_DIR):
        os.makedirs(USER_DIR)
    if not os.path.exists(USER_SCRIPT):
        content = '!/bin/sh\n'
        create_executable(USER_SCRIPT, content)
